#  Copyright (c) Prior Labs GmbH 2025.
#  Licensed under the Apache License, Version 2.0

from __future__ import annotations

import numpy as np
import pandas as pd
import random
import torch
import warnings
from sklearn.base import ClassifierMixin, RegressorMixin
from sklearn.model_selection import train_test_split
from sklearn.preprocessing import LabelEncoder
from sklearn.tree import BaseDecisionTree, DecisionTreeClassifier, DecisionTreeRegressor

from tabpfn_extensions import TabPFNRegressor, TabPFNClassifier


class DecisionTreeTabPFNBase(BaseDecisionTree):
    """Class that implements a DT-TabPFN model based on sklearn package"""

    task_type = None

    def __init__(
        self,
        *,
        # Decision Tree Arguments
        criterion="gini",
        splitter="best",
        max_depth=None,
        min_samples_split=1000,
        min_samples_leaf=1,
        min_weight_fraction_leaf=0.0,
        max_features=None,
        random_state=None,
        max_leaf_nodes=None,
        min_impurity_decrease=0.0,
        class_weight=None,
        ccp_alpha=0.0,
        monotonic_cst=None,
        # TabPFN Arguments
        tabpfn: TabPFNRegressor | TabPFNClassifier = None,
        categorical_features=None,
        verbose=False,
        show_progress=False,
        fit_nodes=True,
        tree_seed=0,
        adaptive_tree=True,
        adaptive_tree_min_train_samples=50,
        adaptive_tree_max_train_samples=2000,
        adaptive_tree_min_valid_samples_fraction_of_train=0.2,
        adaptive_tree_overwrite_metric=None,
        adaptive_tree_test_size: float = 0.2,
        average_logits=True,
        adaptive_tree_skip_class_missing=True,
    ):
        """:param criterion: DecisionTree parameter, see Sklearn documentation
        :param splitter: DecisionTree parameter, see Sklearn documentation
        :param max_depth: DecisionTree parameter, see Sklearn documentation
        :param min_samples_split: DecisionTree parameter, see Sklearn documentation
        :param min_samples_leaf: DecisionTree parameter, see Sklearn documentation
        :param min_weight_fraction_leaf: DecisionTree parameter, see Sklearn documentation
        :param max_features: DecisionTree parameter, see Sklearn documentation
        :param random_state: DecisionTree parameter, see Sklearn documentation
        :param max_leaf_nodes: DecisionTree parameter, see Sklearn documentation
        :param min_impurity_decrease: DecisionTree parameter, see Sklearn documentation
        :param class_weight: DecisionTree parameter, see Sklearn documentation
        :param ccp_alpha: DecisionTree parameter, see Sklearn documentation
        :param monotonic_cst: DecisionTree parameter, see Sklearn documentation
        :param tabpfn: TabPFNBase Model
        :param categorical_features:
        :param verbose:
        :param show_progress:
        :param fit_nodes: Wheather to fit the nodes of the decision tree or only the leafs
        :param tree_seed:
        :param adaptive_tree: Wheather to use adaptive tree, if true, the tree will be pruned based on the performance
            in the out of bag split
        :param adaptive_tree_min_train_samples: TODO: Could we merge this with min_samples_leaf
        :param adaptive_tree_max_train_samples:
        :param adaptive_tree_min_valid_samples_fraction_of_train:
        :param adaptive_tree_overwrite_metric:
        :param adaptive_tree_test_size: Whats the hold out split fraction
        :param average_logits:
        :param adaptive_tree_skip_class_missing:
        """
        # Optional args are arguments that may not be present in BaseDecisionTree, depending on the sklearn version
        # This code helps keep compatability across versions
        optional_args = {"monotonic_cst": monotonic_cst}
        optional_args_filtered = {}
        for key, value in optional_args.items():
            if BaseDecisionTree.__init__.__code__.co_varnames.__contains__(key):
                optional_args_filtered[key] = value

        super().__init__(
            criterion=criterion,
            max_depth=max_depth,
            min_samples_split=min_samples_split,
            min_samples_leaf=min_samples_leaf,
            min_weight_fraction_leaf=min_weight_fraction_leaf,
            max_features=max_features,
            max_leaf_nodes=max_leaf_nodes,
            min_impurity_decrease=min_impurity_decrease,
            random_state=random_state,
            splitter=splitter,
            **optional_args_filtered,
        )

        self.criterion = criterion
        self.max_depth = max_depth
        self.min_samples_leaf = min_samples_leaf
        self.min_weight_fraction_leaf = min_weight_fraction_leaf
        self.max_leaf_nodes = max_leaf_nodes
        self.min_impurity_decrease = min_impurity_decrease
        self.class_weight = class_weight
        self.ccp_alpha = ccp_alpha
        self.monotonic_cst = monotonic_cst
        self.tabpfn = tabpfn
        if self.tabpfn:
            self.tabpfn.random_state = None  # Make sure that TabPFN is not seeded
        self.min_samples_split = min_samples_split
        self.max_features = max_features
        self.random_state = random_state
        self.n_classes_ = 0
        self.tree_seed = tree_seed
        self.classes_ = []
        self.decision_tree = None
        self.leaf_nodes = []
        self.leaf_train_data = {}
        self.categorical_features = categorical_features
        self.verbose = verbose
        self.show_progress = show_progress
        self.splitter = splitter
        self.fit_nodes = fit_nodes
        self.adaptive_tree = adaptive_tree
        self.adaptive_tree_min_train_samples = adaptive_tree_min_train_samples
        self.adaptive_tree_min_valid_samples_fraction_of_train = (
            adaptive_tree_min_valid_samples_fraction_of_train
        )
        self.adaptive_tree_max_train_samples = adaptive_tree_max_train_samples
        self.adaptive_tree_overwrite_metric = adaptive_tree_overwrite_metric
        self.adaptive_tree_test_size = adaptive_tree_test_size
        self.label_encoder_ = LabelEncoder()
        self.average_logits = average_logits
        self.adaptive_tree_skip_class_missing = adaptive_tree_skip_class_missing

    def _more_tags(self):
        return {"multilabel": True, "allow_nan": True}

    def __sklearn_tags__(self):
        tags = super().__sklearn_tags__()
        tags.input_tags.allow_nan = True
        tags.estimator_type = "classifier"
        return tags

    def set_categorical_features(self, categorical_features):
        """Sets categorical features
        :param categorical_features: Categorical features
        :return: None
        """
        self.categorical_features = categorical_features

    def fit(self, X, y, sample_weight=None, check_input=True):
        return self._fit(X, y, sample_weight, check_input=check_input)

    def _fit(
        self,
        X,
        y,
        sample_weight=None,
        check_input=True,
        missing_values_in_feature_mask=None,
    ):
        """Method that fits the DecisionTree-TabPFN model
        :param X: Feature training data
        :param y: Label training data
        :param sample_weight: NOT IMPLEMENTED
        :param check_input:
        :return: None
        """
        self.tree_seed = random.randint(1, 10000)

        if check_input:
            pass
            # X = check_array(X)

        # Handle X and y
        X_preprocessed = self.preprocess_data_for_tree(X)

        sample_weight = (
            np.ones((X_preprocessed.shape[0],))
            if sample_weight is None
            else sample_weight
        )

        y = y.copy()
        self.classes_ = np.unique(y)
        y = pd.DataFrame(y).reset_index().drop("index", axis=1)
        self.n_classes_ = int(y.max().iloc[0] + 1)
        y_train = pd.DataFrame()
        y_train["class"] = y

        # Split data into train and validation set
        if (self.task_type == "multiclass") and (
            np.unique(y, return_counts=True)[1].min() == 1
        ):
            # If there is a class with only one sample, we cannot split the data properly.
            self.adaptive_tree = False

        if self.adaptive_tree:
            stratify = y if self.task_type == "multiclass" else None
            (
                train_X,
                valid_X,
                train_X_preprocessed,
                valid_X_preprocessed,
                train_y,
                valid_y,
                train_sample_weight,
                valid_sample_weight,
            ) = train_test_split(
                X,
                X_preprocessed,
                y,
                sample_weight,
                test_size=self.adaptive_tree_test_size,
                random_state=self.random_state,
                stratify=stratify,
            )

            if self.task_type == "multiclass" and len(np.unique(train_y)) != len(
                np.unique(valid_y),
            ):
                self.adaptive_tree = False
        else:
            train_X, train_X_preprocessed, train_y, train_sample_weight = (
                X,
                X_preprocessed,
                y,
                sample_weight,
            )

        # Create DecisionTree and get leaf nodes of each datapoint
        self._tree = self.init_decision_tree()

        self._tree.fit(
            train_X_preprocessed,
            train_y,
            sample_weight=train_sample_weight,
        )

        if self.verbose:
            self.post_fit()

        (
            self.X,
            self.y,
            self.train_X,
            self.train_X_preprocessed,
            self.train_y,
            self.train_sample_weight,
        ) = (
            X,
            y,
            train_X,
            train_X_preprocessed,
            train_y,
            train_sample_weight,
        )
        if self.adaptive_tree:
            (
                self.valid_X,
                self.valid_X_preprocessed,
                self.valid_y,
                self.valid_sample_weight,
            ) = (
                valid_X,
                valid_X_preprocessed,
                valid_y,
                valid_sample_weight,
            )

        self.todo_post_fit = True

    def fit_leafs(self, train_X, train_y):
        self.leaf_train_data = {}

        # Get leaf nodes of each datapoint in bootstrap sample
        self.leaf_nodes, bootstrap_X, bootstrap_y = self.apply_tree_train(
            train_X,
            train_y,
        )
        if self.verbose:
            print(
                f"Estimators: {self.leaf_nodes.shape[2]}, Nodes per estimator: {self.leaf_nodes.shape[1]}",
            )

        # Store train data point for each leaf
        for estimator_id in range(self.leaf_nodes.shape[2]):
            self.leaf_train_data[estimator_id] = {}
            for leaf_id in range(self.leaf_nodes.shape[1]):
                indices = np.argwhere(self.leaf_nodes[:, leaf_id, estimator_id]).ravel()
                X_train_samples = np.take(train_X, indices, axis=0)
                y_train_samples = np.array(np.take(train_y, indices, axis=0)).ravel()

                if self.verbose and not self.adaptive_tree:
                    print(
                        f"Leaf: {leaf_id} Shape: {X_train_samples.shape} / {train_X.shape}",
                    )

                self.leaf_train_data[estimator_id][leaf_id] = (
                    X_train_samples,
                    y_train_samples,
                )

    def get_tree(self):
        return self._tree

    def apply_tree(self, X):
        """Apply tree for different kinds of tree types.
        TODO: This function could also be overwritten in each type of tree

        (N_samples, N_nodes, N_estimators)
        :param bootstrap_X:
        :return:
        """
        decision_path = self.get_tree().decision_path(X)
        # If trees is a single tree, apply it and expand the missing dimension
        return np.expand_dims(decision_path.todense(), 2)

    def apply_tree_train(self, X, y):
        return self.apply_tree(X), X, y

    def init_decision_tree(self):
        raise NotImplementedError("init_decision_tree must be implemented")

    def post_fit(self):
        pass

    def preprocess_data_for_tree(self, X):
        if torch.is_tensor(X):
            X = X.cpu().numpy()
        X[np.isnan(X)] = 0.0
        return X

    def predict_(self, X, y=None, check_input=True):
        """Predicts X
        :param X: Data that should be evaluated
        :param y: True labels of holdout data used for adaptive tree.
            - If not None: Prunes nodes based on the performance of the holdout data y
            - If None: Predicts the data based on the previous hold out performances
        :param check_input:
        :return: Probabilities of each class
        """
        # TODO: This is a bit of a hack, but it is the only way to stop the fitting at a time budget
        #    We should probably refactor this to a more elegant solution where we move more logic to fit

        if self.todo_post_fit:
            self.todo_post_fit = False
            if self.adaptive_tree:
                if self.verbose:
                    print(
                        f"Starting tree leaf validation train {self.train_X.shape} valid {self.valid_X.shape}",
                    )
                # Pre-fits the leafs of the tree
                self.fit_leafs(self.train_X, self.train_y)
                self.predict_(
                    self.valid_X,
                    self.valid_y.values.ravel(),
                    check_input=False,
                )

            self.fit_leafs(self.X, self.y)

        #### END OF FITTING #####

        X_preprocessed = self.preprocess_data_for_tree(X)
        self.tabpfn.categorical_features_indices = self.categorical_features

        # Get leaf nodes of each datapoint in X
        # Matrix of shape (N_samples, N_nodes, N_estimators) with a one if the sample went trough that decision node
        X_test_leaf_nodes = self.apply_tree(X_preprocessed)
        N_samples, N_nodes, N_estimators = X_test_leaf_nodes.shape

        # We store the probabilities and metrics after evaluating each node or leaf
        y_prob, y_metric = {}, {}
        for estimator_id in range(N_estimators):
            y_prob[estimator_id], y_metric[estimator_id] = {}, {}

        prune_leafs_and_nodes = y is not None and self.adaptive_tree
        if prune_leafs_and_nodes:
            self.node_prediction_type = (
                {}
            )  # Node prediction type stores if a node was pruned or not
            for estimator_id in range(N_estimators):
                self.node_prediction_type[estimator_id] = {}

        if self.show_progress:
            import tqdm.auto

        # Make prediction via TabPFN on each leaf node
        for estimator_id in range(N_estimators):
            loop = (
                tqdm.auto.tqdm(range(N_nodes)) if self.show_progress else range(N_nodes)
            )
            for leaf_id in loop:
                # Initializes the predictions of the node, uses previous predictions
                self._pruning_init_node_predictions(
                    leaf_id,
                    estimator_id,
                    y_prob,
                    y_metric,
                    N_nodes,
                    N_samples,
                )

                # if first leaf of new tree, use last predictions of previous tree, since this is always the same
                if estimator_id > 0 and leaf_id == 0:
                    # print("Skipping leaf 0", estimator_id, leaf_id)
                    continue

                # Fetch the test sample indices that are in the node
                test_sample_indices = np.argwhere(
                    X_test_leaf_nodes[:, leaf_id, estimator_id],
                ).ravel()

                # Fetch the train samples that are in the node
                X_train_samples, y_train_samples = self.leaf_train_data[estimator_id][
                    leaf_id
                ]

                if X_train_samples.shape[0] == 0 or len(test_sample_indices) == 0:
                    if prune_leafs_and_nodes:
                        self.node_prediction_type[estimator_id][leaf_id] = "previous"
                    continue

                is_leaf = (
                    X_test_leaf_nodes[
                        test_sample_indices,
                        leaf_id + 1 :,
                        estimator_id,
                    ].sum()
                    == 0.0
                )  # if larger than 0, then it is not a leaf

                if (
                    not is_leaf
                    and not self.fit_nodes
                    and not (leaf_id == 0 and self.adaptive_tree)
                ):
                    if prune_leafs_and_nodes:
                        self.node_prediction_type[estimator_id][leaf_id] = "previous"
                    continue

                if self.adaptive_tree and leaf_id != 0:
                    # if we are running predictions and the node was skipped
                    if (
                        y is None
                        and self.node_prediction_type[estimator_id][leaf_id]
                        == "previous"
                    ):
                        continue

                    if (
                        self.task_type == "multiclass"
                        and len(np.unique(y_train_samples)) < self.n_classes_
                        and self.adaptive_tree_skip_class_missing
                    ):
                        self.node_prediction_type[estimator_id][leaf_id] = "previous"
                        continue

                    # if not enough train or test datapoints in a node, skip it
                    if (
                        (
                            X_train_samples.shape[0]
                            < self.adaptive_tree_min_train_samples
                        )
                        or (
                            len(test_sample_indices)
                            < self.adaptive_tree_min_valid_samples_fraction_of_train
                            * self.adaptive_tree_min_train_samples
                        )
                        or (
                            X_train_samples.shape[0]
                            > self.adaptive_tree_max_train_samples
                            and not is_leaf
                        )
                    ):
                        if prune_leafs_and_nodes:
                            if self.verbose:
                                print(
                                    f"Leaf: {leaf_id} Tree {estimator_id} No train ({X_train_samples.shape[0]}) / test data {len(test_sample_indices)}",
                                )
                            self.node_prediction_type[estimator_id][
                                leaf_id
                            ] = "previous"

                        continue

                if self.verbose:
                    if y is not None:
                        print(
                            f"Leaf: {leaf_id} Shape: {X_train_samples.shape} Shape test: {len(test_sample_indices)} / {X_preprocessed.shape}"
                            f" N classes (Train Test) {len(np.unique(y_train_samples))} {len(np.unique(y[test_sample_indices]))} /  {self.n_classes_}",
                        )
                    else:
                        print(
                            f"Leaf: {leaf_id} Shape: {X_train_samples.shape} Shape test: {len(test_sample_indices)} / {X_preprocessed.shape}"
                            f" N classes (Train Test) {len(np.unique(y_train_samples))} /  {self.n_classes_}",
                        )

                leaf_prediction = self.predict_leaf_(
                    X_train_samples,
                    y_train_samples,
                    leaf_id,
                    X,
                    test_sample_indices,
                )

                (
                    y_prob_averaging,
                    y_prob_replacement,
                ) = self._pruning_get_prediction_type_results(
                    y_prob,
                    leaf_prediction,
                    test_sample_indices,
                    estimator_id,
                    leaf_id,
                )

                if self.adaptive_tree:
                    if y is not None:
                        self._pruning_set_node_prediction_type(
                            y,
                            y_prob_averaging,
                            y_prob_replacement,
                            y_metric,
                            estimator_id,
                            leaf_id,
                        )
                    self._pruning_set_predictions(
                        y_prob,
                        y_prob_averaging,
                        y_prob_replacement,
                        estimator_id,
                        leaf_id,
                    )
                    if y is not None:
                        y_metric[estimator_id][leaf_id] = self._score(
                            y,
                            y_prob[estimator_id][leaf_id],
                        )
                        if self.verbose:
                            print(
                                f"Leaf: {leaf_id} Score: {y_metric[estimator_id][leaf_id]}",
                            )
                else:
                    y_prob[estimator_id][leaf_id] = y_prob_replacement

        return y_prob[N_estimators - 1][N_nodes - 1]

    def get_optimize_metric(self):
        if self.adaptive_tree_overwrite_metric is not None:
            return self.adaptive_tree_overwrite_metric
        return "roc" if self.task_type == "multiclass" else "rmse"

    def _score(self, y_true, y_pred):
        from ..scoring import scoring_utils

        if self.task_type == "multiclass":
            return scoring_utils.score_classification(
                self.get_optimize_metric(),
                y_true,
                y_pred,
            )
        if self.task_type == "regression":
            return scoring_utils.score_regression(
                self.get_optimize_metric(),
                y_true,
                y_pred,
            )
        return None

    def _pruning_init_node_predictions(
        self,
        leaf_id,
        estimator_id,
        y_prob,
        y_metric,
        N_nodes,
        N_samples,
    ):
        if estimator_id == 0 and leaf_id == 0:
            y_prob[0][0] = self.init_eval_prob(N_samples, to_zero=True)
            y_metric[0][0] = 0
        elif (
            leaf_id == 0 and estimator_id > 0
        ):  # if first leaf of new tree, use last predictions of previous tree
            y_prob[estimator_id][leaf_id] = y_prob[estimator_id - 1][N_nodes - 1]
            y_metric[estimator_id][leaf_id] = y_metric[estimator_id - 1][N_nodes - 1]
        else:  # else use last predictions of previous leaf
            y_prob[estimator_id][leaf_id] = y_prob[estimator_id][leaf_id - 1]
            y_metric[estimator_id][leaf_id] = y_metric[estimator_id][leaf_id - 1]

    @staticmethod
    def softmax(logits):
        exp_logits = np.exp(logits)  # Apply exponential to each logit
        sum_exp_logits = np.sum(
            exp_logits,
            axis=-1,
            keepdims=True,
        )  # Sum of exponentials across classes
        return exp_logits / sum_exp_logits  # Normalize to get probabilities

    def _pruning_get_prediction_type_results(
        self,
        y_eval_prob,
        leaf_prediction,
        test_sample_indices,
        estimator_id,
        leaf_id,
    ):
        y_prob_replacement = np.copy(y_eval_prob[estimator_id][leaf_id])
        y_prob_replacement[test_sample_indices] = leaf_prediction[test_sample_indices]
        if self.task_type == "multiclass":
            y_prob_replacement /= y_prob_replacement.sum(axis=1)[:, None]

        y_prob_averaging = np.copy(y_eval_prob[estimator_id][leaf_id])
        if self.task_type == "multiclass":
            if self.average_logits:
                y_prob_averaging[test_sample_indices] = np.log(
                    y_prob_averaging[test_sample_indices] + 1e-6,
                )
                leaf_prediction[test_sample_indices] = np.log(
                    leaf_prediction[test_sample_indices] + 1e-6,
                )
                # y_prob_averaging[test_sample_indices] += leaf_prediction[test_sample_indices]

                # convert all_proba logits for the multiclass output to probabilities per class
                y_prob_averaging[test_sample_indices] = DecisionTreeTabPFNBase.softmax(
                    y_prob_averaging[test_sample_indices],
                )
            else:
                y_prob_averaging[test_sample_indices] += leaf_prediction[
                    test_sample_indices
                ]
                y_prob_averaging[test_sample_indices] /= y_prob_averaging[
                    test_sample_indices
                ].sum(axis=1)[:, None]
        else:
            y_prob_averaging[test_sample_indices] += leaf_prediction[
                test_sample_indices
            ]
            y_prob_averaging[test_sample_indices] /= 2

        return y_prob_averaging, y_prob_replacement

    def _pruning_set_predictions(
        self,
        y_prob,
        y_prob_averaging,
        y_prob_replacement,
        estimator_id,
        leaf_id,
    ):
        if self.node_prediction_type[estimator_id][leaf_id] == "averaging":
            y_prob[estimator_id][leaf_id] = y_prob_averaging
        elif self.node_prediction_type[estimator_id][leaf_id] == "replacement":
            y_prob[estimator_id][leaf_id] = y_prob_replacement
        else:
            y_prob[estimator_id][leaf_id] = y_prob[estimator_id][leaf_id - 1]

    def _pruning_set_node_prediction_type(
        self,
        y,
        y_prob_averaging,
        y_prob_replacement,
        y_metric,
        estimator_id,
        leaf_id,
    ):
        averaging_score = self._score(y, y_prob_averaging)
        replacement_score = self._score(y, y_prob_replacement)

        if (
            leaf_id == 0
            or max(averaging_score, replacement_score)
            > y_metric[estimator_id][leaf_id - 1]
        ):
            if replacement_score > averaging_score:
                prediction_type = "replacement"
            else:
                prediction_type = "averaging"
        else:
            prediction_type = "previous"

        self.node_prediction_type[estimator_id][leaf_id] = prediction_type

        if self.verbose:
            print(
                f"Leaf: {leaf_id} Averaging: {averaging_score} Replacement: {replacement_score}",
            )

    def predict_leaf_(self, X_train_samples, y_train_samples, leaf_id, X, indices):
        raise NotImplementedError("predict_leaf_ must be implemented")

    def init_eval_prob(self, n_samples, to_zero=False):
        if self.task_type == "multiclass":
            y_eval_prob = np.ones((n_samples, self.n_classes_)) / self.n_classes_
            if to_zero:
                y_eval_prob *= 0
        elif self.task_type == "regression":
            y_eval_prob = np.zeros((n_samples,))
        else:
            raise NotImplementedError("Task type not implemented")
        return y_eval_prob


class DecisionTreeTabPFNClassifier(ClassifierMixin, DecisionTreeTabPFNBase):
    """Class that implements a DT-TabPFN model based on sklearn package"""

    task_type = "multiclass"

    def predict_leaf_(self, X_train_samples, y_train_samples, leaf_id, X, indices):
        y_eval_prob = self.init_eval_prob(X.shape[0], to_zero=True)
        classes = np.array(np.unique(list(map(int, y_train_samples))))

        # Make actual prediction
        if len(classes) > 1:
            try:
                self.tabpfn.random_state = leaf_id + self.tree_seed
                self.tabpfn.fit(X_train_samples, y_train_samples)
                y_eval_prob[indices[:, None], classes] = self.tabpfn.predict_proba(
                    X[indices],
                )
            except ValueError as e:
                if (len(e.args) == 0) or e.args[
                    0
                ] != "All features are constant and would have been removed! Unable to predict using TabPFN.":
                    raise e

                warnings.warn(
                    "One leaf node has bad data to predict with TabPFN (e.g., only constant features). Returning ratio as predictions.",
                    stacklevel=2,
                )
                y_eval_prob[indices[:, None], np.array(classes)] = np.unique(
                    y_train_samples,
                    return_counts=True,
                )[1] / len(y_train_samples)
        else:
            y_eval_prob[indices[:, None], np.array(classes)] = 1

        return y_eval_prob

    def predict(self, X, check_input=True):
        """Predicts X_test
        :param X: Data that should be evaluated
        :param check_input:
        :return: Labels of the predictions
        """
        return np.array(
            list(map(np.argmax, self.predict_proba(X, check_input=check_input))),
        )

    def predict_proba(self, X, check_input=True):
        """Predicts X_test
        :param X: Data that should be evaluated
        :param check_input:
        :return: Probabilities of each class
        """
        return self.predict_(X, check_input=check_input)

    def post_fit(self):
        pass

    def init_decision_tree(self):
        return DecisionTreeClassifier(
            criterion=self.criterion,
            max_depth=self.max_depth,
            min_samples_split=self.min_samples_split,
            min_samples_leaf=self.min_samples_leaf,
            min_weight_fraction_leaf=self.min_weight_fraction_leaf,
            max_features=self.max_features,
            random_state=self.random_state,
            max_leaf_nodes=self.max_leaf_nodes,
            min_impurity_decrease=self.min_impurity_decrease,
            class_weight=self.class_weight,
            ccp_alpha=self.ccp_alpha,
            splitter=self.splitter,
        )


class DecisionTreeTabPFNRegressor(RegressorMixin, DecisionTreeTabPFNBase):
    """Class that implements a DT-TabPFN model based on sklearn package"""

    task_type = "regression"

    def __init__(
        self,
        *,
        # Decision Tree Arguments
        criterion="squared_error",
        splitter="best",
        max_depth=None,
        min_samples_split=1000,
        min_samples_leaf=1,
        min_weight_fraction_leaf=0.0,
        max_features=None,
        random_state=None,
        max_leaf_nodes=None,
        min_impurity_decrease=0.0,
        ccp_alpha=0.0,
        monotonic_cst=None,
        # TabPFN Arguments
        tabpfn=None,
        categorical_features=None,
        verbose=False,
        show_progress=False,
        fit_nodes=True,
        tree_seed=0,
        adaptive_tree=True,
        adaptive_tree_min_train_samples=50,
        adaptive_tree_max_train_samples=2000,
        adaptive_tree_min_valid_samples_fraction_of_train=0.2,
        adaptive_tree_overwrite_metric=None,
        adaptive_tree_test_size=0.2,
        average_logits=True,
        adaptive_tree_skip_class_missing=True,
    ):
        super().__init__(
            criterion=criterion,
            splitter=splitter,
            max_depth=max_depth,
            min_samples_split=min_samples_split,
            min_samples_leaf=min_samples_leaf,
            min_weight_fraction_leaf=min_weight_fraction_leaf,
            max_features=max_features,
            random_state=random_state,
            max_leaf_nodes=max_leaf_nodes,
            min_impurity_decrease=min_impurity_decrease,
            ccp_alpha=ccp_alpha,
            monotonic_cst=monotonic_cst,
            tabpfn=tabpfn,
            categorical_features=categorical_features,
            verbose=verbose,
            show_progress=show_progress,
            fit_nodes=fit_nodes,
            tree_seed=tree_seed,
            adaptive_tree=adaptive_tree,
            adaptive_tree_min_train_samples=adaptive_tree_min_train_samples,
            adaptive_tree_max_train_samples=adaptive_tree_max_train_samples,
            adaptive_tree_min_valid_samples_fraction_of_train=adaptive_tree_min_valid_samples_fraction_of_train,
            adaptive_tree_overwrite_metric=adaptive_tree_overwrite_metric,
            adaptive_tree_test_size=adaptive_tree_test_size,
            average_logits=average_logits,
            adaptive_tree_skip_class_missing=adaptive_tree_skip_class_missing,
        )

    def __sklearn_tags__(self):
        tags = super().__sklearn_tags__()
        tags.input_tags.allow_nan = True
        tags.estimator_type = "regressor"
        return tags

    def init_decision_tree(self):
        return DecisionTreeRegressor(
            criterion=self.criterion,
            max_depth=self.max_depth,
            min_samples_split=self.min_samples_split,
            min_samples_leaf=self.min_samples_leaf,
            min_weight_fraction_leaf=self.min_weight_fraction_leaf,
            max_features=self.max_features,
            random_state=self.random_state,
            max_leaf_nodes=self.max_leaf_nodes,
            min_impurity_decrease=self.min_impurity_decrease,
            ccp_alpha=self.ccp_alpha,
            splitter=self.splitter,
        )

    def predict_leaf_(self, X_train_samples, y_train_samples, leaf_id, X, indices):
        self.tabpfn.random_state = leaf_id + self.tree_seed
        self.tabpfn.fit(X_train_samples, y_train_samples)
        y_eval_prob = np.zeros((X.shape[0],))
        y_eval_prob[indices] = self.tabpfn.predict(X[indices])

        return y_eval_prob

    def predict(self, X, check_input=True):
        """Predicts X_test
        :param X: Data that should be evaluated
        :param check_input:
        :return: Labels of the predictions
        """
        return self.predict_(X, check_input=check_input)

    def predict_full(self, X):
        """Predicts X
        :param X: Data that should be evaluated
        :param check_input:
        :return: Labels of the predictions
        """
        return self.predict_(X, check_input=False)

    def post_fit(self):
        pass
        # print(f"Decision Tree fitted with N nodes: {self._tree.tree_.node_count}")



#  Copyright (c) Prior Labs GmbH 2025.
#  Licensed under the Apache License, Version 2.0

from __future__ import annotations

import logging
import numpy as np
import pandas as pd
import warnings
from sklearn.model_selection import KFold, RepeatedStratifiedKFold, StratifiedKFold

logger = logging.getLogger(__name__)


def _save_stratified_splits(
    _splitter: StratifiedKFold | RepeatedStratifiedKFold,
    x: np.ndarray,
    y: np.ndarray,
    n_splits: int,
    auto_fix_stratified_splits: bool = False,
) -> list[list[list[int], list[int]]]:
    """Fix from AutoGluon to avoid unsafe splits for classification if less than n_splits instances exist for all classes.

    https://github.com/autogluon/autogluon/blob/0ab001a1193869a88f7af846723d23245781a1ac/core/src/autogluon/core/utils/utils.py#L70.
    """
    try:
        splits = [
            [train_index, test_index]
            for train_index, test_index in _splitter.split(x, y)
        ]
    except ValueError as e:
        x = pd.DataFrame(x)
        y = pd.Series(y)
        y_dummy = pd.concat([y, pd.Series([-1] * n_splits)], ignore_index=True)
        X_dummy = pd.concat([x, x.head(n_splits)], ignore_index=True)
        invalid_index = set(y_dummy.tail(n_splits).index)
        splits = [
            [train_index, test_index]
            for train_index, test_index in _splitter.split(X_dummy, y_dummy)
        ]
        len_out = len(splits)
        for i in range(len_out):
            train_index, test_index = splits[i]
            splits[i][0] = [
                index for index in train_index if index not in invalid_index
            ]
            splits[i][1] = [index for index in test_index if index not in invalid_index]

        # only rais afterward because only now we know that we cannot fix it
        if not auto_fix_stratified_splits:
            raise AssertionError(
                "Cannot split data in a stratifed way with each class in each subset of the data.",
            ) from e
    except UserWarning as e:
        # Assume UserWarning for not enough classes for correct stratified splitting.
        raise e

    return splits


def fix_split_by_dropping_classes(
    x: np.ndarray,
    y: np.ndarray,
    n_splits: int,
    spliter_kwargs: dict,
) -> list[list[list[int], list[int]]]:
    """Fixes stratifed splits for edge case.

    For each class that has fewer instances than number of splits, we oversample before split to n_splits and then remove all oversamples and
    original samples from the splits; effectively removing the class from the data without touching the indices.
    """
    val, counts = np.unique(y, return_counts=True)
    too_low = val[counts < n_splits]
    too_low_counts = counts[counts < n_splits]

    y_dummy = pd.Series(y.copy())
    X_dummy = pd.DataFrame(x.copy())
    org_index_max = len(X_dummy)
    invalid_index = []

    for c_val, c_count in zip(too_low, too_low_counts, strict=True):
        fill_missing = n_splits - c_count
        invalid_index.extend(np.where(y == c_val)[0])
        y_dummy = pd.concat(
            [y_dummy, pd.Series([c_val] * fill_missing)],
            ignore_index=True,
        )
        X_dummy = pd.concat(
            [X_dummy, pd.DataFrame(x).head(fill_missing)],
            ignore_index=True,
        )

    invalid_index.extend(list(range(org_index_max, len(y_dummy))))
    splits = _save_stratified_splits(
        _splitter=StratifiedKFold(**spliter_kwargs),
        x=X_dummy,
        y=y_dummy,
        n_splits=n_splits,
    )
    len_out = len(splits)
    for i in range(len_out):
        train_index, test_index = splits[i]
        splits[i][0] = [index for index in train_index if index not in invalid_index]
        splits[i][1] = [index for index in test_index if index not in invalid_index]

    return splits


def assert_valid_splits(
    splits: list[list[list[int], list[int]]],
    y: np.ndarray,
    *,
    non_empty: bool = True,
    each_selected_class_in_each_split_subset: bool = True,
    same_length_training_splits: bool = True,
):
    """Verify that the splits are valid."""
    if non_empty:
        assert len(splits) != 0, "No splits generated!"
        for split in splits:
            assert len(split) != 0, "Some split is empty!"
            assert len(split[0]) != 0, "A train subset of a split is empty!"
            assert len(split[1]) != 0, "A test subset of a split is empty!"

    if each_selected_class_in_each_split_subset:
        # As we might drop classes, we first need to build the set of classes that are in the splits.
        #   - 2nd unique is for speed up purposes only.
        _real_y = set(
            np.unique([c for split in splits for c in np.unique(y[split[1]])]),
        )
        # Now we need to check that each class that exists in all splits is in each split.
        for split in splits:
            assert _real_y == (
                set(np.unique(y[split[0]]))
            ), "A class is missing in a train subset!"
            assert _real_y == (
                set(np.unique(y[split[1]]))
            ), "A class is missing in a test subset!"

    if same_length_training_splits:
        for split in splits:
            assert len(split[0]) == len(
                splits[0][0],
            ), "A train split has different amount of samples!"


def _equalize_training_splits(
    input_splits: list[list[list[int], list[int]]],
    rng: np.random.RandomState,
) -> list[list[list[int], list[int]]]:
    """Equalize training splits by duplicating samples in too small splits."""
    splits = input_splits[:]
    n_max_train_samples = max(len(split[0]) for split in splits)
    for split in splits:
        curr_split_len = len(split[0])
        if curr_split_len < n_max_train_samples:
            missing_samples = n_max_train_samples - curr_split_len
            split[0].extend(
                [int(dup_i) for dup_i in rng.choice(split[0], size=missing_samples)],
            )
            split[0] = sorted(split[0])

    return splits


def get_cv_split_for_data(
    x: np.ndarray,
    y: np.ndarray,
    splits_seed: int,
    n_splits: int,
    *,
    stratified_split: bool,
    safety_shuffle: bool = True,
    auto_fix_stratified_splits: bool = False,
    force_same_length_training_splits: bool = False,
) -> list[list[list[int], list[int]]] | str:
    """Safety shuffle and generate (safe) splits.

    If it returns str at the first entry, no valid split could be generated and the str is the reason why.
    Due to the safety shuffle, the original x and y are also returned and must be used.

    Note: the function does not support repeated splits at this point.
    Simply call this function multiple times with different seeds to get repeated splits.

    Test with:

    ```python
        if __name__ == "__main__":
        print(
            get_cv_split_for_data(
                x=np.array([[1, 2, 3, 4, 5, 6, 7, 8, 9, 10], [1, 2, 3, 4, 5, 6, 7, 8, 9, 10]]).T,
                y=np.array([1, 1, 1, 2, 2, 2, 3, 3, 3, 4]),
                splits_seed=42,
                n_splits=3,
                stratified_split=True,
                auto_fix_stratified_splits=True,
            )
        )
    ```

    Args:
        x: The data to split.
        y: The labels to split.
        splits_seed: The seed to use for the splits. Or a RandomState object.
        n_splits: The number of splits to generate.
        stratified_split: Whether to use stratified splits.
        safety_shuffle: Whether to shuffle the data before splitting.
        auto_fix_stratified_splits: Whether to try to fix stratified splits automatically.
            Fix by dropping classes with less than n_splits samples.
        force_same_length_training_splits: Whether to force the training splits to have the same amount of samples.
            Force by duplicating random instance in the training subset of a too small split until all training splits have the same amount of samples.
    Out:
        A list of pairs of indexes, where in each pair first come the train examples, then test. So we get something like
        `[[TRAIN_INDICES_0, TEST_INDICES_0], [TRAIN_INDICES_1, TRAIN_INDICES_1]]` for 2 splits.
        Or a string if no valid split could be generated whereby the string gives the reason.
    """
    assert len(x) == len(y), "x and y must have the same length!"

    rng = np.random.RandomState(splits_seed)
    if safety_shuffle:
        p = rng.permutation(len(x))
        x, y = x[p], y[p]
    spliter_kwargs = {"n_splits": n_splits, "shuffle": True, "random_state": rng}

    if not stratified_split:
        splits = [list(tpl) for tpl in KFold(**spliter_kwargs).split(x, y)]
    else:
        with warnings.catch_warnings():
            warnings.filterwarnings("error")
            try:
                splits = _save_stratified_splits(
                    _splitter=StratifiedKFold(**spliter_kwargs),
                    x=x,
                    y=y,
                    n_splits=n_splits,
                    auto_fix_stratified_splits=auto_fix_stratified_splits,
                )
            except UserWarning as e:
                logger.debug(e)
                if auto_fix_stratified_splits:
                    logger.debug("Trying to fix stratified splits automatically...")
                    splits = fix_split_by_dropping_classes(
                        x=x,
                        y=y,
                        n_splits=n_splits,
                        spliter_kwargs=spliter_kwargs,
                    )
                else:
                    splits = "Cannot generate valid stratified splits for dataset without losing classes in some subsets!"
            except AssertionError as e:
                logger.debug(e)
                splits = "Cannot generate valid stratified splits for dataset without losing classes in some subsets!"

    if isinstance(splits, str):
        return splits

    if force_same_length_training_splits:
        splits = _equalize_training_splits(splits, rng)

    assert_valid_splits(
        splits=splits,
        y=y,
        non_empty=True,
        same_length_training_splits=force_same_length_training_splits,
        each_selected_class_in_each_split_subset=stratified_split,
    )

    if safety_shuffle:
        # Revert to correct outer scope indices
        for split in splits:
            split[0] = sorted(p[split[0]])
            split[1] = sorted(p[split[1]])

    return splits



#  Copyright (c) Prior Labs GmbH 2025.
#  Licensed under the Apache License, Version 2.0

from __future__ import annotations

import logging
import numpy as np
import time
from abc import ABC, abstractmethod
from collections.abc import Callable
from copy import deepcopy
from sklearn.base import BaseEstimator, ClassifierMixin, RegressorMixin
from sklearn.model_selection import ShuffleSplit, StratifiedShuffleSplit
from typing import Literal

from .save_splitting import get_cv_split_for_data
from ..scoring.scoring_utils import (
    CLF_LABEL_METRICS,
    score_classification,
    score_regression,
)

logging.basicConfig(
    format="%(asctime)s %(levelname)-8s %(message)s",
    level=logging.INFO,
    datefmt="%Y-%m-%d %H:%M:%S",
)
logger = logging.getLogger(__name__)


class AbstractValidationUtils(ABC, BaseEstimator):
    _holdout_index_hit_list: np.ndarray | None = (
        None  # list of indices that were hit during repeated holdout validation
    )

    def __init__(
        self,
        estimators: list[tuple[str, BaseEstimator]],
        n_repeats: int,
        seed: int,
        *,
        holdout_fraction: float = 0.33,
        validation_method: Literal["cv", "holdout"] = "cv",
        n_folds: int | None = None,
        time_limit: int | None = None,
        score_metric: str | None = None,
    ):
        """Abstract validation utilities for sklearn-like models.

        Args:
            estimators: list of tuples (name, estimator)
            n_repeats: number of repeats for getting OOF
            seed: random seed
            time_limit: If None, no time limit is set. Otherwise, this specifics time limit in seconds for early stopping cross-validation of the models.
                If the time limit set, the number of repeats is are the maximum amount of repeats. As a result, this code does not guarantee that to stop before
                the time_limit is over.
            score_metric: metric to use for scoring. If None, child class defines default on the fly.
            validation_method: method to use for validation. If "cv", (repeated) cross-validation is used. If "holdout", (repeated) holdout validation is used.
            n_folds: number of folds for getting OOF. Only used if validation_method is "cv"
            holdout_fraction: fraction of data to hold out for holdout validation. Only used if validation_method is "holdout"
        """
        self.estimators = estimators
        self.score_metric = score_metric
        self.n_folds = n_folds
        self.n_repeats = n_repeats
        self.seed = seed
        self._rng = np.random.RandomState(seed)
        self.time_limit = time_limit
        self._start_time: int = 0
        self.classes_: np.ndarray | None = None
        self._repeats_seed = [
            int(self._rng.randint(0, int(np.iinfo(np.int32).max))) for _ in range(self.n_repeats)
        ]
        self._estimators = estimators.copy()  # internal copy for early stopping.
        self.validation_method = validation_method
        self.holdout_fraction = holdout_fraction

    @property
    @abstractmethod
    def stratifed_split(self) -> bool:
        pass

    @abstractmethod
    def loss_function(self, y_true, y_pred_proba) -> float:
        pass

    @abstractmethod
    def _predict_oof(self, base_model, X) -> np.ndarray:
        pass

    @abstractmethod
    def _batch_predict_oof(self, X, configurations) -> np.ndarray:
        pass

    @abstractmethod
    def _proba_template(self, X, y) -> np.ndarray:
        pass

    @property
    def _is_classification(self) -> bool:
        return isinstance(self, AbstractValidationUtilsClassification)

    @property
    def _is_holdout(self) -> bool:
        return self.validation_method == "holdout"

    def _fill_predictions_in_place(
        self,
        *,
        model_i: int,
        base_model: Callable,
        oof_proba_list: list[np.ndarray],
        X: np.ndarray,
        y: np.ndarray,
        train_index: list[int],
        test_index: list[int],
        loss_per_estimator: list[list[float]] | None,
        holdout_index_hits: np.ndarray | None,
        _extra_processing: bool,
        split_i: int,
    ) -> None:
        fold_X_train, fold_X_test = X[train_index], X[test_index]
        fold_y_train, fold_y_test = y[train_index], y[test_index]
        # base_model = copy.deepcopy(base_model) # FIXME: think about adding this for safety but will likely slow down (due to having to load model again)

        # Default base models case
        base_model.fit(fold_X_train, fold_y_train)

        pred = self._predict_oof(base_model, fold_X_test)

        oof_proba_list[model_i][test_index] += pred

        if holdout_index_hits is not None:
            holdout_index_hits[test_index] += 1

        if loss_per_estimator is not None:
            loss_per_estimator[model_i].append(
                self.loss_function(fold_y_test, pred),
            )
        if _extra_processing:  # custom code for supporting extra use case (ignore)
            if not hasattr(self, "_extra_processing"):
                raise ValueError(
                    f"_extra_processing is True, but _extra_processing is not defined for the current class: {type(self)}.",
                )
            self._extra_processing(split_i, base_model, test_index)

    def _get_split(
        self,
        *,
        X,
        y,
        fold_i: int,
        repeat_i: int,
        holdout_validation: bool,
    ) -> list[list[int], list[int]]:
        if holdout_validation:
            _spliter = StratifiedShuffleSplit if self.stratifed_split else ShuffleSplit

            return list(
                _spliter(
                    n_splits=self.n_repeats,
                    test_size=self.holdout_fraction,
                    random_state=self._repeats_seed[0],  # take the first one.
                ).split(X, y),
            )[repeat_i]

        return get_cv_split_for_data(
            x=X,
            y=y,
            stratified_split=self.stratifed_split,
            n_splits=self.n_folds,
            splits_seed=self._repeats_seed[repeat_i],
            safety_shuffle=True,
            auto_fix_stratified_splits=True,
        )[fold_i]

    def yield_cv_data_iterator(
        self,
        X,
        y,
    ) -> tuple[int, int, int, BaseEstimator, list[int], list[int], bool, bool,]:
        n_models = len(self.estimators)
        holdout_validation = self._is_holdout
        _folds = self.n_folds if not holdout_validation else 1

        for repeat_i in range(self.n_repeats):
            for model_i in range(n_models):
                for fold_i in range(_folds):
                    split_i = fold_i + repeat_i * _folds
                    train_index, test_index = _get_split = self._get_split(
                        X=X,
                        y=y,
                        fold_i=fold_i,
                        repeat_i=repeat_i,
                        holdout_validation=holdout_validation,
                    )
                    check_for_model_early_stopping = False
                    check_for_repeat_early_stopping = False

                    if (fold_i + 1) == _folds:  # check after last fold of model
                        check_for_model_early_stopping = True
                        if (
                            model_i + 1
                        ) == n_models:  # check after last fold of last model
                            check_for_repeat_early_stopping = True

                    logger.info(
                        f"Yield data for model {self.estimators[model_i][0]} and split {split_i} (repeat={repeat_i + 1}).",
                    )
                    yield (
                        model_i,
                        split_i,
                        repeat_i + 1,
                        self.estimators[model_i][1],
                        train_index,
                        test_index,
                        check_for_model_early_stopping,
                        check_for_repeat_early_stopping,
                    )

    def time_limit_reached(self) -> bool:
        if self.time_limit is None:
            return False
        return (time.time() - self._start_time) > self.time_limit

    def not_enough_time(self, current_repeat: int) -> bool:
        """Simple heuristic to stop cross-validation early if not enough time is left for another repeat."""
        if self.time_limit is None:
            return False

        time_spent_so_far = time.time() - self._start_time
        avg_time_per_repeat = time_spent_so_far / current_repeat

        return (time_spent_so_far + avg_time_per_repeat) > self.time_limit

    def set_time_limit(self) -> None:
        if self.time_limit is not None:
            self._start_time = time.time()
            logger.info(
                f"Set time limit to {self.time_limit} seconds. We will early stop validation if needed.",
            )

    def _impute_dropped_instances(
        self,
        *,
        proba_template: np.ndarray,
        oof_proba_list: list[np.ndarray],
        y: np.ndarray,
    ) -> list[np.ndarray]:
        fill_val = (
            np.mean(y) if proba_template.shape[1] == 1 else 1 / proba_template.shape[1]
        )

        for oof in oof_proba_list:
            # Only impute if any values were not filled (i.e., are NaN)
            if np.isnan(oof).any():
                logger.warning(
                    "Imputing instances that were dropped during splits (e.g., due to not enough instances per class).",
                )
                oof[np.isnan(oof).any(axis=1)] = fill_val

        # Fix dropped classes
        classes_ = np.unique(y)
        if len(classes_) != self.n_classes_:
            logger.warning(
                "Adding classes to OOF proba that were dropped during splits (e.g., due to not enough instances per class).",
            )
            missing_classes = np.setdiff1d(classes_, self.classes_)

            new_oof = []
            for oof in oof_proba_list:
                tmp_oof = np.full((oof.shape[0], len(classes_)), np.nan)
                for i, c in enumerate(classes_):
                    if c in missing_classes:
                        tmp_oof[:, i] = 0
                    else:
                        tmp_oof[:, i] = oof[
                            :,
                            np.where(self.classes_ == c)[0],
                        ].flatten()
                new_oof.append(tmp_oof)
            oof_proba_list = new_oof

        return oof_proba_list

    def get_oof_per_estimator(
        self,
        X: np.ndarray,
        y: np.ndarray,
        *,
        return_loss_per_estimator: bool = False,
        impute_dropped_instances: bool = True,
        _extra_processing: bool = False,
    ) -> list[np.ndarray] | tuple[list[np.ndarray], list[float]]:
        """Get OOF predictions for each base model.

        Args:
            X: training data (features)
            y: training labels
            return_loss_per_estimator: if True, also return the loss per estimator.
            impute_dropped_instances: if True, impute instances that were dropped during the splits (e.g., due to not enough instances per class).
            _extra_processing:

        Returns: either only OOF predictions or OOF predictions and loss per estimator.
            If self.is_holdout is True, the OOF predictions can return NaN values for instances not covered during repeated holdout.
        """
        if self.validation_method == "cv":
            if self.n_folds is None:
                raise ValueError(
                    "Number of folds must be specified for cross-validation.",
                )
            logger.info(
                f"Starting {self.n_repeats}-repeated {self.n_folds}-fold cross-validation.",
            )
        elif self.validation_method == "holdout":
            if not isinstance(self.holdout_fraction, float):
                raise ValueError(
                    f"Holdout fraction must be specified as a float for holdout validation. Got: {self.holdout_fraction}",
                )
            logger.info(
                f"Starting {self.n_repeats}-repeated holdout validation with holdout_frac={self.holdout_fraction}.",
            )
        else:
            raise ValueError(f"Unknown validation method: {self.validation_method}")

        (
            oof_proba_list,
            backup_oof_proba_list,
            backup_loss_per_estimator,
            backup_run_repeats,
        ) = (None, None, None, None)
        use_backups = False  # only use backups if at least on repeat was done.
        loss_per_estimator = (
            [[] for _ in self.estimators] if return_loss_per_estimator else None
        )
        holdout_index_hits, holdout_index_hits_backup, holdout_index_hit_counts = (
            None,
            None,
            None,
        )

        # Get OOF predictions for each base model
        self.set_time_limit()
        ran_repeats = self.n_repeats

        for (
            model_i,
            split_i,
            current_repeat,
            base_model,
            train_index,
            test_index,
            check_for_model_early_stopping,
            check_for_repeat_early_stopping,
        ) in self.yield_cv_data_iterator(X, y):
            # Delayed setting classes_ to make sure that fixes for splits have been applied.
            if (self.classes_ is None) and self._is_classification:
                self.classes_ = np.unique(y[train_index])

            if oof_proba_list is None:
                proba_template = self._proba_template(X, y)
                oof_proba_list = [deepcopy(proba_template) for _ in self.estimators]

            to_pass_holdout_index_hits = None
            if (holdout_index_hits is None) and self._is_holdout:
                holdout_index_hits = np.zeros(X.shape[0], dtype=float)
                holdout_index_hit_counts = 0
            if self._is_holdout and holdout_index_hit_counts < current_repeat:
                to_pass_holdout_index_hits = holdout_index_hits
                holdout_index_hit_counts = current_repeat

            self._fill_predictions_in_place(
                model_i=model_i,
                base_model=base_model,
                oof_proba_list=oof_proba_list,
                X=X,
                y=y,
                train_index=train_index,
                test_index=test_index,
                loss_per_estimator=loss_per_estimator,
                holdout_index_hits=to_pass_holdout_index_hits,
                split_i=split_i,
                _extra_processing=_extra_processing,
            )

            if check_for_repeat_early_stopping:  # True after every repeat.
                ran_repeats = current_repeat

                # Save backups for model early stopping
                backup_oof_proba_list = deepcopy(oof_proba_list)
                backup_run_repeats = ran_repeats
                backup_loss_per_estimator = (
                    None if loss_per_estimator is None else deepcopy(loss_per_estimator)
                )
                holdout_index_hits_backup = (
                    None if holdout_index_hits is None else deepcopy(holdout_index_hits)
                )
                use_backups = True  # use backups after first repeat

                logger.info(
                    f"Finished validation repeat {current_repeat}/{self.n_repeats}.",
                )
                if self.time_limit_reached():
                    logger.info("Time limit reached. Stop repeating validation.")
                    break

                if self.not_enough_time(current_repeat=current_repeat):
                    logger.info(
                        "Likely not enough time left for another repeat. Stop repeating validation.",
                    )
                    break

            if check_for_model_early_stopping:
                if self.time_limit_reached():
                    logger.info("Time limit reached.")
                elif self.not_enough_time(
                    current_repeat=current_repeat * (model_i + 1),
                ):
                    logger.info("Likely not enough time left for another model.")
                else:
                    continue
                logger.info(
                    f"Stop validation of all models after {model_i + 1} models in repeat {current_repeat}.",
                )

                if use_backups:
                    logger.info("Use results from last repeat for all models!")
                    # Fallback to last repeat if time limit is reached and we did cv before.
                    oof_proba_list = backup_oof_proba_list
                    loss_per_estimator = backup_loss_per_estimator
                    holdout_index_hits = holdout_index_hits_backup
                    ran_repeats = backup_run_repeats
                else:
                    logger.info(
                        "As this is the first repeat, we trim down the models to all so-far run models!",
                    )
                    # Trim down OOF and losses to only have validated models.
                    ran_repeats = current_repeat
                    oof_proba_list = [
                        oof_proba_list[tmp_i] for tmp_i in range(model_i + 1)
                    ]
                    loss_per_estimator = (
                        [loss_per_estimator[tmp_i] for tmp_i in range(model_i + 1)]
                        if loss_per_estimator is not None
                        else None
                    )
                    self._estimators = self._estimators[: model_i + 1]
                break

        # Set empty OOF predictions to NaN
        for i, oof in enumerate(oof_proba_list):
            if self._is_classification:
                nan_mask = oof.sum(axis=1) == 0
            elif self._is_holdout:
                nan_mask = holdout_index_hits == 0
            else:
                # Regression without holdout. No need to set to NaN as all values must be filled (no dropped instances, no holdout).
                break
            oof_proba_list[i][nan_mask] = np.nan

        if self._is_holdout:
            self._holdout_index_hit_list = np.where(holdout_index_hits > 0)[0]

        # Sanity check
        #  - check that all OOF predictions sum up to the number of repeats for classification
        if self._is_classification:
            if self._is_holdout:
                holdout_index_hits[holdout_index_hits == 0] = np.nan
                if not all(
                    np.isclose(
                        oof.sum(axis=1),
                        holdout_index_hits,
                        equal_nan=True,
                    ).all()
                    for oof in oof_proba_list
                ):
                    raise ValueError(
                        "OOF predictions are not consistent over repeats for holdout! Something went wrong.",
                    )
            elif not all(
                np.isclose(
                    oof[~np.isnan(oof).any(axis=1)].sum(axis=1),
                    ran_repeats,
                ).all()
                for oof in oof_proba_list
            ):
                for i, oof in enumerate(oof_proba_list):
                    print("==================== ", i)  # STUPID DEBUG PRINT
                    print(
                        oof[
                            ~np.isclose(
                                oof[~np.isnan(oof).any(axis=1)].sum(axis=1),
                                ran_repeats,
                            )
                        ],
                    )
                    print(ran_repeats)
                print()
                raise ValueError(
                    "OOF predictions are not consistent over repeats! Something went wrong.",
                )

        # Average OOF predictions over all repeats
        if ran_repeats > 1:
            for i, oof in enumerate(oof_proba_list):
                div_by = (
                    ran_repeats
                    if not self._is_holdout
                    else holdout_index_hits[:, None]
                    if self._is_classification
                    else holdout_index_hits
                )
                oof_proba_list[i] = oof / div_by

        # Imputation of dropped instances is only needed for stratified splits as we do not drop values otherwise.
        if impute_dropped_instances and self.stratifed_split and (not self._is_holdout):
            oof_proba_list = self._impute_dropped_instances(
                proba_template=proba_template,
                oof_proba_list=oof_proba_list,
                y=y,
            )

        if loss_per_estimator is not None:
            loss_per_estimator = [np.mean(losses) for losses in loss_per_estimator]
            return oof_proba_list, loss_per_estimator

        return oof_proba_list


class AbstractValidationUtilsClassification(AbstractValidationUtils, ClassifierMixin):
    n_classes_: int

    def loss_function(self, y_true, y_pred_proba):
        opt_metric = "log_loss" if self.score_metric is None else self.score_metric

        y_pred_is_labels = False
        if opt_metric in CLF_LABEL_METRICS:
            y_pred_proba = self.classes_.take(np.argmax(y_pred_proba, axis=1))
            y_pred_is_labels = True

        # negative to make score a loss
        return -score_classification(
            opt_metric,
            y_true,
            y_pred_proba,
            y_pred_is_labels=y_pred_is_labels,
        )

    def _batch_predict_oof(self, X, configurations):
        return self.tabpfn_model.custom_batch_predict_proba(
            X=X,
            configurations=configurations,
        )

    def _predict_oof(self, base_model, X):
        return base_model.predict_proba(X)

    @property
    def stratifed_split(self):
        return True

    def _proba_template(self, X, y):
        self.n_classes_ = len(self.classes_)
        return np.full((X.shape[0], self.n_classes_), 0, dtype=float)


class AbstractValidationUtilsRegression(AbstractValidationUtils, RegressorMixin):
    def loss_function(self, y_true, y_pred_proba):
        opt_metric = "rmse" if self.score_metric is None else self.score_metric
        return -score_regression(opt_metric, y_true, y_pred_proba)

    def _predict_oof(self, base_model, X):
        return base_model.predict(X)

    def _batch_predict_oof(self, X, configurations):
        return self.tabpfn_model.custom_batch_predict(
            X=X,
            configurations=configurations,
        )

    @property
    def stratifed_split(self):
        return False

    def _proba_template(self, X, y):
        return np.full((X.shape[0],), 0, dtype=float)
